#include <vector>
#include <MOO-EALib/PopulationMOO.h>

using std::vector;

#include "TAGSNP.h"

void calcMOFitness( IndividualMOO &Ind,
                    bool **cover_table,
                    const BinMat &data,
                    const uint16_t uni_hap_num,
                    const uint16_t SNP_NUM,
                    vector<double> &FV  )
{
    double  f1(.0), f2(.0), f3(.0), f4(.0);
    uint16_t uni_hap_pairs  =   uni_hap_num*(uni_hap_num-1)/2;
    uint16_t raw_hap_pairs  =   HAP_NUM*(HAP_NUM-1)/2;

    uint16_t i, j, k;

    vector<bool> chrom = dynamic_cast< vector< bool >& >(Ind[0]);
    uint16_t * covered = new uint16_t [ uni_hap_pairs ];

    for ( i = SNP_NUM; i--; )   {   f1+=chrom[i];   }

    if ( f1 < ceil(log(uni_hap_num)/log(2)) )   f2 = 0;
    else
    {
        f2 = SNP_NUM;
        for ( j = 0; j < uni_hap_pairs ; j++ )
        {
            covered[j] = 0;

            for ( i = chrom.size(); i--; )
            {
                if (chrom[i])
                    covered[j] += cover_table[i][j];
            }

            if (covered[j] == 0)        //covered[j] = 0.
            {
                f2 = 0;
                break;
            }
            if (covered[j] < f2)    //find minimum covered[j]
                f2 = covered[j];
        }
    }

    if (!f2)
    {
        dynamic_cast< ChromosomeT< bool >& >(Ind[ 0 ]).flip(0.9545);

        f1 = SNP_NUM - f1;

        if ( f1 < ceil(log(uni_hap_num)/log(2)) )   f2 = 0;
        else
        {
            f2 = SNP_NUM;
            for (uint16_t j = 0; j < uni_hap_pairs ; j++)
            {
                covered[j] = 0;

                for (uint16_t i = 0; i < chrom.size(); i++)
                {
                    if (chrom[i])
                        covered[j] += cover_table[i][j];
                }

                if (covered[j] == 0)        //covered[j] = 0.
                {
                    f2 = 0;
                    break;
                }
                if (covered[j] < f2)    //find minimum covered[j]
                    f2 = covered[j];
            }
        }
    }


    if (f2 == 0)
    {
        Ind.setMOOFitnessValues(inf, inf, inf, inf);
        FV[0] = inf;    FV[1] = inf;    FV[2] = inf;    FV[3] = inf;
    }
    else
    {
        if (uni_hap_num != HAP_NUM)
        {
            uint16_t idx_count = 0;
            vector< uint16_t > ham_dist (raw_hap_pairs ,0);

            for (i = 0; i < HAP_NUM-1; i++)
            {
                for (j = i+1; j < HAP_NUM; j++)
                {
                    ham_dist[idx_count] = 0;
                    for ( k = chrom.size(); k--; )
                    {
                        if (chrom[k])
                            ham_dist[idx_count] += (data[i][k] ^ data[j][k]);
                    }
                    f3 += ham_dist[idx_count];
                    f4 += ham_dist[idx_count]*ham_dist[idx_count];
                    idx_count++;
                }
            }
        }
        else
        {
            for ( k = raw_hap_pairs; k--; )
            {
                f3 += covered[k];
                f4 += covered[k] * covered[k];
            }
        }

//        non-normalize
//        f3 /= raw_hap_pairs ;
//        f4 = f4/raw_hap_pairs - (f3/raw_hap_pairs)*(f3/raw_hap_pairs);

//        f4 = sqrt( f4*raw_hap_pairs/(f3*f3) - 1 );      //!"coefficient of variation" of ham-dist among Haplotype pairs
        f3 /= raw_hap_pairs * f1;
        f4 = (f4/(f1*f1))/raw_hap_pairs - f3*f3;

//        f1 /= SNP_NUM;
        f2 = 1/f2;
        f3 = 1/f3;

        Ind.setMOOFitnessValues(f1 , f2 , f3 , f4);
        FV[0] = f1/SNP_NUM;    FV[1] = f2;    FV[2] = f3/(HAP_NUM/2.0);    FV[3] = f4;
    }

    delete[] covered;
}
