//#include <omp.h>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <vector>
#include <time.h>

//#include <EALib/Population.h>
#include <MOO-EALib/PopulationMOO.h>
#include <MOO-EALib/ArchiveMOO.h>

using std::vector;
using std::ofstream;
using std::ios;
using std::setw;
using std::cout;
using std::endl;
using std::max;
using std::sort;

#include "TAGSNP.h"

static void Aggregation (const  vector< vector<double> > &wv, const vector< vector<double> > &fv,
                            vector< vector<uint16_t> > &rank)
{
    uint16_t Psize = fv.size();
    uint16_t TVsize = wv.size();

    register uint16_t i, j;

//    double magnitude(.0), cos_theta(.0);

    vector< vector<double> > score_and_rank (Psize, vector<double> (2, .0));

//    vector< double > best_ind_weight (TVsize, inf);
//    vector< bool > improvement_mask (TVsize, false);

    for (i = rank.size(); i--;)     rank[i].back() = i;

    for (j = TVsize; j--;) //!Target Vector
    {
//        if( j < TVsize )
            for( i = Psize; i--; )
            {   //! Calculate Weighted MinMax and store into score_and_rank vector
                score_and_rank[i].front() = Rng::gauss(0.0, 1.0)*1e-10 +
                max(
                    max(
                        max(
                            fv[i][0]/(wv[j][0]+eps),
                            fv[i][1]/(wv[j][1]+eps) ),
                        fv[i][2]/(wv[j][2]+eps) ),
                    fv[i][3]/(wv[j][3]+eps) );

                score_and_rank[i].back() = i;
            }
//        else
//            for( i = Psize; i--; )
//            {   //! Calculate Vector Angle Distance Scaling and store into score_and_rank vector
//                magnitude = sqrt(   fv[i][0]*fv[i][0]+
//                                    fv[i][1]*fv[i][1]+
//                                    fv[i][2]*fv[i][2]+
//                                    fv[i][3]*fv[i][3]   );
//                cos_theta =     (   fv[i][0]*wv[j][0]+
//                                    fv[i][1]*wv[j][1]+
//                                    fv[i][2]*wv[j][2]+
//                                    fv[i][3]*wv[j][3]   )/magnitude;
//
//                score_and_rank[i].front() = magnitude/(pow(cos_theta, TVsize) + 0.01) + Rng::gauss(0.0, 1.0)*1e-10;
//                score_and_rank[i].back() = i;
//            }

        sort(score_and_rank.begin(), score_and_rank.end());

        //! Assign rank value to each individual
        for( i = Psize; i--; )     rank[ score_and_rank[i].back() ][j] = i;

        //! Record the best individual found so far
//        if ( score_and_rank.front().front() <= best_ind_weight[j] )
//        {
//            best_ind_weight[j] = score_and_rank.front().front();
//            improvement_mask[j] = true;
//        }
    }

    for( i = Psize; i--; )  sort(rank[i].begin(), rank[i].end()-1);
    sort(rank.begin(), rank.end());
}

static vector<uint16_t> UniquePop (PopulationMOO &p)
{
    uint16_t psize( p.size() ), i(0) ,j(0);
    vector<uint16_t> dup;

    for(i = 0; i < psize-1; i++)
        for(j = i+1; j < psize; j++)
            if ( p[i][0] == p[j][0] )
            {
                dynamic_cast< ChromosomeT< bool >& >(p[ j ][ 0 ]).initialize();
                dup.push_back(j);
            }

    return dup;
}

bool Rank_comp(Individual*const& pInd1, Individual*const& pInd2)
{
	IndividualMOO* pI1 = (IndividualMOO*)pInd1;
	IndividualMOO* pI2 = (IndividualMOO*)pInd2;

	unsigned Rank1 = pI1->getMOORank();
	unsigned Rank2 = pI2->getMOORank();

	return (Rank1 < Rank2);
}

bool V_comp(vector< uint16_t > a, vector< uint16_t > b)
{
    a.pop_back();
    b.pop_back();

    return (a == b);
}

static void GreedyInit(uint16_t SNP_NUM, uint16_t pairs_num, const BinMat &data, bool **cover_table,BinMat &InitMatrix)
{
    uint16_t i, j, SNP_idx;

    vector< bool > first_cover (pairs_num, false);
    vector< bool > all_covered (pairs_num, true);

    vector< vector<uint16_t> > arranged_data (SNP_NUM, vector<uint16_t>(1 + HAP_NUM + 1, 0) );
    vector< vector<uint16_t> >::iterator it;

    for (i = 0;i < SNP_NUM;i++)
    {
        uint16_t snp_sum = 0;
        for (j = 0;j < HAP_NUM;j++)
        {
            snp_sum += data[j][i];
            arranged_data[i][1+j] = data[j][i];
        }
        arranged_data[i].front() = HAP_NUM/2 - abs(snp_sum - HAP_NUM/2);
        arranged_data[i].back() = i;

//        InitMatrix[2][i] = true;
    }

    random_shuffle(arranged_data.begin(), arranged_data.end());
    sort(arranged_data.begin(), arranged_data.end());
    it = unique(arranged_data.begin(), arranged_data.end(), V_comp);
    arranged_data.resize( it - arranged_data.begin() );

    for (i = arranged_data.size()-1; first_cover != all_covered; i--)
    {
        SNP_idx = arranged_data[i].back();
        for (j = pairs_num;j--; )
        {
            first_cover[j] = first_cover[j] | cover_table[ SNP_idx ][j];
        }

        InitMatrix[0][ SNP_idx ] = true;
    }

    for (i = arranged_data.size(); i--;)
    {
        InitMatrix[1][ arranged_data[i].back() ] = true;
    }

//    cout << arranged_data.size() << "\n";
//    for (i = 0;i<first_cover.size();i++)
//    {
//            cout << first_cover[i] << "\t";
//    }
//    cout << "\n";
}

static void Identical(uint16_t SNP_NUM, uint16_t Pairs_NUM, bool **cover_table, vector< vector<uint16_t> > &ident)
{
    register uint16_t i, j;

    vector< uint16_t > push_v;
    vector< vector<uint16_t> > rearranged_data (SNP_NUM, vector<uint16_t>(Pairs_NUM + 1, 0) );

    for ( i = SNP_NUM; i--; )
    {
        std::copy(cover_table[i], cover_table[i]+Pairs_NUM, rearranged_data[i].begin());

        rearranged_data[i].back() = i;
    }

    sort(rearranged_data.begin(), rearranged_data.end());

    push_v.push_back( rearranged_data.front().back() );
    for ( i = 0; i < SNP_NUM-1; i++ )
    {
        if ( V_comp(rearranged_data[i], rearranged_data[i+1]) )
        {
            push_v.push_back( rearranged_data[i+1].back() );
        }
        else
        {
            ident.push_back( push_v );
            push_v.clear();
            push_v.push_back( rearranged_data[i+1].back() );
        }
    }
    ident.push_back( push_v );
}

static ChromosomeT< bool > Unique_init(const uint16_t SNP_NUM, const vector< vector<uint16_t> > &Ident)
{
    ChromosomeT< bool > chr(SNP_NUM, false);
//    double pr(.0);
    uint16_t idx(0);

	for (uint16_t i = Ident.size(); i--; )
	{
	    idx = Rng::discrete(0, Ident[i].size());

        if ( idx < Ident[i].size() )
            chr[ Ident[i][idx] ] = true;
//	    pr = 1.0/(1 + Ident[i].size());
//	    for (uint16_t j = Ident[i].size(); j--; )
//	    {
//	        chr[ Ident[i][j] ] = bool( Rng::coinToss( pr ) );
//	    }
	}

    return chr;
}

static ChromosomeT< bool > Greedy_init(const uint16_t SNP_NUM, const uint16_t Pairs_NUM, bool **cover_table, vector< vector<uint16_t> > ident)
{
    ChromosomeT< bool > chr(SNP_NUM, false);
    uint16_t idx(0);

    vector< bool > all_covered (Pairs_NUM, true);
    vector< bool > merged_covered (Pairs_NUM, false);

    random_shuffle(ident.begin(), ident.end());

	for (uint16_t i = ident.size()-1; merged_covered != all_covered; i--)
	{
	    idx = Rng::discrete(0, ident[i].size()-1);

	    chr[ ident[i][idx] ] = true;

        for ( uint16_t j = Pairs_NUM; j--; )
            merged_covered[j] = merged_covered[j] | cover_table[ ident[i][idx] ][j];
	}

    return chr;
}


static inline void Output_Pop(ofstream &out_pop, const uint16_t PopSize, PopulationMOO &p)
{
    register uint16_t i;
    uint16_t dim = p[0][0].size();
    vector< vector<double> > FitnessVectors (PopSize, vector<double> (4, 0.0));

    for (i = PopSize; i--;)
        FitnessVectors[i] = p[i].getMOOFitnessValues();

    for (i = PopSize; i--;)
        out_pop << FitnessVectors[i][0] << ",";
    out_pop << endl;
    for (i = PopSize; i--;)
        out_pop << 1 / FitnessVectors[i][1] << ",";
    out_pop << endl;
    for (i = PopSize; i--;)
        out_pop << 1 / FitnessVectors[i][2] << ",";
    out_pop << endl;
    for (i = PopSize; i--;)
        out_pop << FitnessVectors[i][3] << ",";
    out_pop << endl;

    out_pop << endl;
}

static void Output_BlockPareto(ArchiveMOO &archive, uint16_t snp_num, uint16_t BlockNo, uint16_t psize, double g)
{
    char filename[50];
    sprintf(filename, "P%03d_G%03d%%-ParetoFinal.csv", psize, int(g*100+0.5));
    ofstream block_pareto (filename, ios::app);

    vector<double> FitnessVector (4, .0);

    for (uint16_t i = 0; i < archive.size(); ++i)
    {
        FitnessVector = (archive.readArchive(i)).getMOOFitnessValues();

        block_pareto << setw(3) << BlockNo << ","
        <<      FitnessVector[0] << ","
        << 1 /  FitnessVector[1] << ","
        << 1 /  FitnessVector[2] << ","
        <<      FitnessVector[3] << "\n";
    }
    block_pareto.close();
}

static void Output_SelectedSNP(uint16_t SNP_NUM, const BinMat &data, ArchiveMOO &archive, uint16_t BlockNo)
{
    char filename[50];
    sprintf(filename, "\\ans\\TagSol_%04d.txt", BlockNo);
    ofstream ansout(filename);

    vector< bool > sol_chrom (SNP_NUM, false);

    for (uint16_t i = 0; i < archive.size(); ++i)
    {
        sol_chrom = dynamic_cast< vector< bool >& >(archive.readArchive(i)[0]);
        for (uint16_t j = 0; j < data.size(); ++j)
        {
            for (uint16_t k = 0; k < SNP_NUM; ++k)
            {
                if( sol_chrom[k] )
                    ansout << data[j][k];
                else
                    ansout << "-";
            }
            ansout << "\n";
        }
        ansout << "\n";
    }
    ansout.close();
}

void NSGAII(bool **cover_table, const BinMat &data, const vector< vector<double> > &WeightV,
            uint16_t uni_hap_num, const uint16_t SNP_NUM, const uint16_t BlockNo,
            /*const uint16_t P_SIZE, const uint16_t Generation,*/ const double GreedIsGood)
{
    const uint16_t  PopSize     =   200;
    const uint16_t  Dimension   =   SNP_NUM;
    const uint16_t  Iterations  =   500;
    const double    Xrate       =   0.7;
    const double    Frate       =   1.0 / Dimension;
    const uint32_t  archiveSize =   std::numeric_limits< unsigned int >::max();
    const uint16_t  No_Obj      =   4;
    const uint16_t  No_TarVec   =   WeightV.size();
//    const double    GreedIsGood =   1/3.0;

    int i;
    uint16_t a, b;
    uint16_t NO_nondom;
    clock_t t1,t2;
//    vector< uint16_t > duplicate;

    PopulationMOO parents(PopSize, ChromosomeT< bool >(Dimension));
    PopulationMOO offsprings(PopSize, ChromosomeT< bool >(Dimension));
    PopulationMOO mu_lambda(PopSize*2, ChromosomeT< bool >(Dimension));

    parents   .setMinimize();
    offsprings.setMinimize();
    mu_lambda .setMinimize();

    parents   .setNoOfObj(No_Obj);
    offsprings.setNoOfObj(No_Obj);
    mu_lambda .setNoOfObj(No_Obj);

    ArchiveMOO archive(archiveSize);
    archive.minimize();


    vector< vector<double> >    FitnessV    (PopSize, vector<double>  (No_Obj, .0));
    vector< vector<uint16_t> >  Rank        (PopSize, vector<uint16_t>(No_TarVec+1, 0));

    BinMat InitMatrix (2, vector<bool> (SNP_NUM, false) );
    GreedyInit(SNP_NUM, uni_hap_num*(uni_hap_num-1)/2, data, cover_table, InitMatrix);

    vector< vector<uint16_t> > identical_cover;
    Identical(SNP_NUM, uni_hap_num*(uni_hap_num-1)/2, cover_table, identical_cover);

    Rng::seed( time(NULL) );
    srand ( time(NULL) );

    for (i = 0; i < ceil(GreedIsGood*10); i+=2)
    {
        dynamic_cast< ChromosomeT< bool >& >(parents[ i ][ 0 ]) = InitMatrix[0];
        dynamic_cast< ChromosomeT< bool >& >(parents[i+1][ 0 ]) = InitMatrix[1];
    }
    for (i = ceil(GreedIsGood*10); i < ceil(PopSize*GreedIsGood); i+=2)
    {
        dynamic_cast< ChromosomeT< bool >& >(parents[ i ][ 0 ]) = Greedy_init(SNP_NUM, uni_hap_num*(uni_hap_num-1)/2, cover_table, identical_cover);
        dynamic_cast< ChromosomeT< bool >& >(parents[i+1][ 0 ]) = Unique_init(SNP_NUM, identical_cover);
    }
    for (i = ceil(PopSize*GreedIsGood); i < PopSize; i++)
        dynamic_cast< ChromosomeT< bool >& >(parents[i][ 0 ]).initialize();

    for (i = PopSize; i--;)
        calcMOFitness(parents[ i ], cover_table, data, uni_hap_num, SNP_NUM, FitnessV[i]);


    Aggregation(WeightV, FitnessV, Rank);

    //! pseudo rank value, the smaller the better in the population
    for (i = Rank.size(); i--;)   parents[ Rank[i].back() ].setMOORank( i );

    FitnessV.resize ( PopSize*2, vector<double>  (No_Obj, .0) );
    Rank.resize     ( PopSize*2, vector<uint16_t>  (No_TarVec+1,  0) );

    sort(parents.begin(), parents.end(), Rank_comp);

    char filename[50];
    sprintf(filename, "P%03d_G%03d_%02d-Generation.csv", PopSize, int(GreedIsGood*100+0.5), BlockNo);
//    sprintf(filename, "Generation#%03d.csv", BlockNo);
    ofstream out_pop(filename);

    for (uint16_t it = 1; it <= Iterations; ++it)
    {

        cout << it << "\r";

/*        uint16_t offs_count = 0;
        for (i = 0; i < PopSize; i++)
        {
            if (parents[i].getMOORank() == 1)
            {
                if (offs_count < PopSize/2)
                    offsprings[ offs_count*2 ][ 0 ] = parents[ i ][ 0 ];
                else
                    offsprings[2*(PopSize - offs_count) - 1][ 0 ] = parents[ i ][ 0 ];
//                    offsprings[ offs_count*2 - PopSize + 1][ 0 ] = parents[ i ][ 0 ];

                offs_count++;
            }
            else
                break;
        }

        while ( offs_count < PopSize )
        {
            a = Rng::discrete(0, PopSize-1);
            b = Rng::discrete(0, PopSize-1);

            //! if b better than a
            if( PopulationMOO::compareRankShare(&parents[b], &parents[a]) ) a = b;

            if (offs_count < PopSize/2)
                offsprings[ offs_count*2 ][ 0 ] = parents[ a ][ 0 ];
            else
                offsprings[2*(PopSize - offs_count) - 1][ 0 ] = parents[ a ][ 0 ];
//                offsprings[ offs_count*2 - PopSize + 1][ 0 ] = parents[ a ][ 0 ];

            offs_count++;
        }
*/

        for (i = PopSize; i--;)
        {
            a = Rng::discrete(0, PopSize-1);
            b = Rng::discrete(0, PopSize-1);
            if ( parents[a].getMOORank() < parents[b].getMOORank() )
                offsprings[i][0] = parents[a][0];
            else
                offsprings[i][0] = parents[b][0];
        }

        for (i = 0; i < PopSize; i += 2)
            if (Rng::coinToss(Xrate))
                offsprings[ i ][ 0 ].crossoverUniform( offsprings[ i+1 ][ 0 ] );

        for (i = PopSize; i--;)
            dynamic_cast< ChromosomeT< bool >& >(offsprings[ i ][ 0 ]).flip(Frate);

        for (i = PopSize; i--;)
            calcMOFitness(offsprings[ i ], cover_table, data, uni_hap_num, SNP_NUM, FitnessV[i+PopSize]);


        mu_lambda.combinePopulationMOO(parents, offsprings);

//        duplicate = UniquePop(mu_lambda);
//        for (i = 0; i < duplicate.size(); i ++)
//            calcMOFitness(mu_lambda[ duplicate[i] ], cover_table, data, uni_hap_num, SNP_NUM, FitnessV[ duplicate[i] ]);

        Aggregation(WeightV, FitnessV, Rank);

        for (i = Rank.size(); i--;)     mu_lambda[ Rank[i].back() ].setMOORank( i );

        sort( mu_lambda.begin(), mu_lambda.end(), Rank_comp);

        Output_Pop(out_pop, PopSize*2, mu_lambda);

        for (i = PopSize*2; i--;)
            archive.addArchive(mu_lambda[ i ]);
        archive.nonDominatedSolutions();

        for (i = PopSize; i--;)
        {
            parents[i] = mu_lambda[i];
            FitnessV[i] = parents[i].getMOOFitnessValues();

            FitnessV[i][0] /= SNP_NUM;
            FitnessV[i][2] /= HAP_NUM/2.0;
        }


    }
    out_pop.close();

    Output_BlockPareto(archive, SNP_NUM, BlockNo, PopSize, GreedIsGood);

//    Output_SelectedSNP(SNP_NUM, data, archive, BlockNo);
}
