#include <vector>
#include <MOO-EALib/PopulationMOO.h>

using std::vector;

#include "TAGSNP.h"

void calcMOFitness(IndividualMOO &Ind, bool **cover_table, const BinMat &data,const uint16_t uni_hap_num, const uint16_t SNP_NUM)
{
    double  f1(.0), f2(.0), f3(.0), f4(.0);
    uint16_t uni_hap_pairs  =   uni_hap_num*(uni_hap_num-1)/2;
    uint16_t raw_hap_pairs  =   HAP_NUM*(HAP_NUM-1)/2;

    vector<bool> chrom = dynamic_cast< vector< bool >& >(Ind[0]);
//    vector< uint16_t > covered( cover_mtx[0].size() );
    uint16_t * covered = new uint16_t [ uni_hap_pairs ];

//    BinMat uni_data;

    for (uint16_t i = 0; i < SNP_NUM; i++)  f1+=chrom[i];

    if ( f1 < ceil(log(uni_hap_num)/log(2)) )   f2 = 0;
    else
    {
        f2 = SNP_NUM;
        for (uint16_t j = 0; j < uni_hap_pairs ; j++)
        {
            covered[j] = 0;

            for (uint16_t i = 0; i < chrom.size(); i++)
            {
                if (chrom[i])
                    covered[j] += cover_table[i][j];
            }

            if (covered[j] == 0)        //covered[j] = 0.
            {
                f2 = 0;
                break;
            }
            if (covered[j] < f2)    //find minimum covered[j]
                f2 = covered[j];
        }
    }

    if (!f2)
    {
        dynamic_cast< ChromosomeT< bool >& >(Ind[ 0 ]).flip(1.0);

        f1 = SNP_NUM - f1;

        if ( f1 < ceil(log(uni_hap_num)/log(2)) )   f2 = 0;
        else
        {
            f2 = SNP_NUM;
            for (uint16_t j = 0; j < uni_hap_pairs ; j++)
            {
                covered[j] = 0;

                for (uint16_t i = 0; i < chrom.size(); i++)
                {
                    if (chrom[i])
                        covered[j] += cover_table[i][j];
                }

                if (covered[j] == 0)        //covered[j] = 0.
                {
                    f2 = 0;
                    break;
                }
                if (covered[j] < f2)    //find minimum covered[j]
                    f2 = covered[j];
            }
        }
    }


    if (f2 == 0)
        Ind.setMOOFitnessValues(inf, inf, inf, inf);
    else
    {
        if (uni_hap_num != HAP_NUM)
        {
            uint16_t idx_count = 0;
            vector< uint16_t > ham_dist (raw_hap_pairs ,0);

            for (uint16_t a = 0; a < HAP_NUM-1; a++)
            {
                for (uint16_t b = a+1; b < HAP_NUM; b++)
                {
                    ham_dist[idx_count] = 0;
                    for (uint16_t c = 0; c < chrom.size(); c++)
                    {
                        if (chrom[c])
                            ham_dist[idx_count] += (data[a][c] != data[b][c]);
                    }
                    f3 += ham_dist[idx_count];
                    f4 += ham_dist[idx_count]*ham_dist[idx_count];
                    idx_count++;
                }
            }
        }
        else
        {
            for (uint16_t k = raw_hap_pairs; k--; )
            {
                f3 += covered[k];
                f4 += covered[k] * covered[k];
            }
        }

//        non-normalize
//        f3 /= raw_hap_pairs ;
//        f4 = f4/raw_hap_pairs - (f3/raw_hap_pairs)*(f3/raw_hap_pairs);

//        f4 = sqrt( f4*raw_hap_pairs/(f3*f3) - 1 );      //!"coefficient of variation" of ham-dist among Haplotype pairs
        f3 /= raw_hap_pairs * f1;
        f4 = (f4/(f1*f1))/raw_hap_pairs - f3*f3;

//        f1 /= SNP_NUM;
//        f2 = 1/f2;
//        f3 = 1/f3;

        Ind.setMOOFitnessValues(f1 , 1/f2 ,  1/f3 ,f4);
    }

    delete[] covered;
}
